from autotx import models
from autotx.utils.ethereum import SafeManager
from autotx.wallets.smart_wallet import SmartWallet


class SafeSmartWallet(SmartWallet):
    manager: SafeManager
    auto_submit_tx: bool

    def __init__(self, manager: SafeManager, auto_submit_tx: bool):
        super().__init__(manager.client.w3, manager.address)

        self.manager = manager
        self.auto_submit_tx = auto_submit_tx

    def on_transactions_prepared(self, txs: list[models.Transaction]) -> None:
        pass

    def on_transactions_ready(self, txs: list[models.Transaction]) -> bool | str:
        return self.manager.send_tx_batch(txs, not self.auto_submit_tx)



from enum import Enum
from pydantic import BaseModel, Field
from typing import Any, List, Optional, Union
from datetime import datetime

class TransactionType(str, Enum):
    SEND = "send"
    APPROVE = "approve"
    SWAP = "swap"

class TransactionBase(BaseModel):
    type: TransactionType
    id: str = Field(default="")
    task_id: str = Field(default="")
    summary: str
    params: dict[str, Any]

class SendTransaction(TransactionBase):
    receiver: str
    token_symbol: str
    token_address: str
    amount: float

    @classmethod
    def create(cls, token_symbol: str, token_address: str, amount: float, receiver: str, params: dict[str, Any]) -> 'SendTransaction':
        return cls(
            type=TransactionType.SEND,
            token_symbol=token_symbol,
            token_address=token_address,
            amount=amount,
            receiver=receiver,
            params=params,
            summary=f"Transfer {amount} {token_symbol} to {receiver}",
        )

class ApproveTransaction(TransactionBase):
    token_symbol: str
    token_address: str
    amount: float
    spender: str

    @classmethod
    def create(cls, token_symbol: str, token_address: str, amount: float, spender: str, params: dict[str, Any]) -> 'ApproveTransaction':
        return cls(
            type=TransactionType.APPROVE,
            token_symbol=token_symbol,
            token_address=token_address,
            amount=amount,
            spender=spender,
            params=params,
            summary=f"Approve {amount} {token_symbol} to {spender}"
        )

class SwapTransaction(TransactionBase):
    from_token_symbol: str
    to_token_symbol: str
    from_token_address: str
    to_token_address: str
    from_amount: float
    to_amount: float

    @classmethod
    def create(cls, from_token_symbol: str, to_token_symbol: str, from_token_address: str, to_token_address: str, from_amount: float, to_amount: float, params: dict[str, Any]) -> 'SwapTransaction':
        return cls(
            type=TransactionType.SWAP,
            from_token_symbol=from_token_symbol,
            to_token_symbol=to_token_symbol,
            from_token_address=from_token_address,
            to_token_address=to_token_address,
            from_amount=from_amount,
            to_amount=to_amount,
            params=params,
            summary=f"Swap {from_amount} {from_token_symbol} for at least {to_amount} {to_token_symbol}"
        )

Transaction = Union[SendTransaction, ApproveTransaction, SwapTransaction]

class Task(BaseModel):
    id: str
    prompt: str
    address: str
    chain_id: int
    created_at: datetime
    updated_at: datetime
    error: str | None
    running: bool
    messages: List[str]
    transactions: List[Transaction]

class App(BaseModel):
    id: str
    name: str
    api_key: str
    allowed: bool

class AppUser(BaseModel):
    id: str
    user_id: str
    agent_address: str
    created_at: datetime
    app_id: str

class ConnectionCreate(BaseModel):
    user_id: str

class TaskCreate(BaseModel):
    prompt: str
    address: Optional[str] = None
    chain_id: Optional[int] = None
    user_id: str

class SendTransactionsParams(BaseModel):
    address: str
    chain_id: int
    user_id: str

class SupportedNetwork(BaseModel):
    name: str
    chain_id: int


from dataclasses import dataclass
from decimal import Decimal
from typing import Any, cast

from web3 import Web3
from autotx import models
from autotx.utils.ethereum.constants import GAS_PRICE_MULTIPLIER, NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.erc20_abi import ERC20_ABI
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.helpers.get_native_token_symbol import (
    get_native_token_symbol,
)
from autotx.utils.ethereum.lifi import Lifi, TokenNotSupported
from autotx.utils.ethereum.networks import ChainId
from web3.types import TxParams, Wei

SLIPPAGE = 0.005  # 0.5%

SUPPORTED_NETWORKS_BY_LIFI = [
    ChainId.MAINNET,
    ChainId.OPTIMISM,
    ChainId.ARBITRUM_ONE,
    ChainId.BASE_MAINNET,
    ChainId.POLYGON,
    ChainId.GNOSIS,
    ChainId.ZKSYNC_V2
]

@dataclass
class QuoteInformation:
    approval_address: str
    amount_in: int
    to_amount_min: int
    transaction: TxParams
    exchange_name: str


def get_quote(
    token_in_address: ETHAddress,
    token_in_decimals: int,
    token_in_symbol: str,
    token_out_address: ETHAddress,
    token_out_decimals: int,
    token_out_symbol: str,
    chain: ChainId,
    expected_amount: Decimal,
    amount_is_output: bool,
    from_address: ETHAddress,
) -> QuoteInformation:
    quote: dict[str, Any] | None = None
    try:
        if amount_is_output:
            quote = Lifi.get_quote_to_amount(
                token_in_address,
                token_out_address,
                int(expected_amount * (10**token_out_decimals)),
                from_address,
                chain,
                SLIPPAGE,
            )
            amount_in_integer = int(quote["estimate"]["fromAmount"])

        else:
            amount_in_integer = int(expected_amount * (10**token_in_decimals))
            quote = Lifi.get_quote_from_amount(
                token_in_address,
                token_out_address,
                amount_in_integer,
                from_address,
                chain,
                SLIPPAGE,
            )
    except TokenNotSupported as e:
        if e.token_address.lower() == token_in_address.hex.lower():
            raise Exception(
                f"Token {token_in_symbol} is not supported. Please try another one"
            )
        if e.token_address.lower() == token_out_address.hex.lower():
            raise Exception(
                f"Token {token_out_symbol} is not supported. Please try another one"
            )

    if not quote:
        raise Exception("Quote has not been fetched")

    transaction = TxParams(
        {
            "to": quote["transactionRequest"]["to"],
            "from": quote["transactionRequest"]["from"],
            "data": quote["transactionRequest"]["data"],
            "gasPrice": quote["transactionRequest"]["gasPrice"],
            "gas": quote["transactionRequest"]["gasLimit"],
            "value": Wei(int(quote["transactionRequest"]["value"], 0)),
            "chainId": quote["transactionRequest"]["chainId"]
        }
    )
    return QuoteInformation(
        quote["estimate"]["approvalAddress"],
        amount_in_integer,
        int(quote["estimate"]["toAmountMin"]),
        transaction,
        quote["toolDetails"]["name"],
    )


def build_swap_transaction(
    web3: Web3,
    amount: Decimal,
    token_in_address: ETHAddress,
    token_out_address: ETHAddress,
    _from: ETHAddress,
    is_exact_input: bool,
    chain: ChainId,
) -> list[models.Transaction]:
    native_token_symbol = get_native_token_symbol(chain)
    token_in_is_native = token_in_address.hex == NATIVE_TOKEN_ADDRESS
    token_in = web3.eth.contract(
        address=token_in_address.hex, abi=ERC20_ABI
    )
    token_in_symbol = (
        native_token_symbol
        if token_in_is_native
        else token_in.functions.symbol().call()
    )
    token_in_decimals = (
        18 if token_in_is_native else token_in.functions.decimals().call()
    )

    token_out_is_native = token_out_address.hex == NATIVE_TOKEN_ADDRESS
    token_out = web3.eth.contract(
        address=token_out_address.hex, abi=ERC20_ABI
    )
    token_out_decimals = (
        18 if token_out_is_native else token_out.functions.decimals().call()
    )

    token_out_symbol = (
        native_token_symbol
        if token_out_is_native
        else token_out.functions.symbol().call()
    )
    quote = get_quote(
        token_in_address,
        token_in_decimals,
        token_in_symbol,
        token_out_address,
        token_out_decimals,
        token_out_symbol,
        chain,
        amount,
        not is_exact_input,
        _from,
    )

    transactions: list[models.Transaction] = []
    if not token_in_is_native:
        approval_address = quote.approval_address
        allowance = token_in.functions.allowance(_from.hex, approval_address).call()
        if allowance < quote.amount_in:
            tx = token_in.functions.approve(
                approval_address, quote.amount_in
            ).build_transaction(
                {
                    "from": _from.hex,
                    "gasPrice": Wei(
                        int(web3.eth.gas_price * GAS_PRICE_MULTIPLIER)
                    ),
                }
            )
            transactions.append(
                models.ApproveTransaction.create(
                    token_symbol=token_in_symbol,
                    token_address=str(token_in_address),
                    amount=float(Decimal(str(quote.amount_in)) / 10 ** token_in_decimals),
                    spender=approval_address,
                    params=cast(dict[str, Any], tx),
                )
            )
    transactions.append(
        models.SwapTransaction.create(
            from_token_symbol=token_in_symbol,
            to_token_symbol=token_out_symbol,
            from_token_address=str(token_in_address),
            to_token_address=str(token_out_address),
            from_amount=float(Decimal(str(quote.amount_in)) / 10 ** token_in_decimals),
            to_amount=float(Decimal(str(int(quote.to_amount_min))) / 10 ** token_out_decimals),
            params=cast(dict[str, Any], quote.transaction),
        )
    )
    return transactions
